import os

from bentoml.exceptions import (
    InvalidArgument,
    MissingDependencyException,
)
from bentoml.service import BentoServiceArtifact
import torch

try:
    from simpletransformers.classification import MultiLabelClassificationModel
except ImportError:
    MultiLabelClassificationModel = None


class SimpleTransformersModelArtifact(BentoServiceArtifact):

    def __init__(self,
        name,
        model_type='roberta',
        num_labels=None,
        pos_weight=None,
        args=None,
        use_cuda=False,
        cuda_device=-1,
        **kwargs
    ):
        super(SimpleTransformersModelArtifact, self).__init__(name)
        self._model = None
        self._model_type = model_type
        self._num_labels = num_labels
        self._pos_weight = pos_weight
        self._args = args
        self._use_cuda = use_cuda
        self._cuda_device = cuda_device
        self._kwargs = kwargs

        if MultiLabelClassificationModel is None:
            raise MissingDependencyException(
                'simpletransformers.classification.MultiLabelClassificationModel is required to use SimpleTransformersModelArtifact'
            )

    def _file_path(self, base_path):
        return os.path.join(base_path, self.name)

    def _load_from_directory(self, path):
        self._model = MultiLabelClassificationModel(
            self._model_type,
            self._file_path(path),
            num_labels=self._num_labels,
            pos_weight=self._pos_weight,
            args=self._args,
            use_cuda=self._use_cuda,
            cuda_device=self._cuda_device,
            **self._kwargs
        )
    
    def pack(self, model):
        if isinstance(model, str):
            if os.path.isdir(model):
                self._load_from_directory(model)
                return self
        
        raise InvalidArgument('Expecting a path to the model directory')


    def load(self, path):
        return self.pack(path)

    def save(self, dst):
        path = self._file_path(dst)
        os.makedirs(path, exist_ok=True)
        model = self._model.model
        model_to_save = model.module if hasattr(model, 'module') else model
        model_to_save.save_pretrained(path)
        self._model.tokenizer.save_pretrained(path)
        torch.save(self._model.args, os.path.join(path, 'training_args.bin'))
        return path

    def get(self):
        return self._model



import os
import tempfile

import click
import docker
import stringcase
from bentoml.utils.lazy_loader import LazyLoader
from jinja2 import Environment, FileSystemLoader
from kubernetes import client, config

import from_yaml from bentoutils.kubeutil


@click.command()
@click.option('--mod', help='fully qualified module name containing service to package')
@click.option('--clz', help='class name of service to package')
@click.option('--name', help='model name')
@click.option('--path', help='directory path of pretrained model')
def pack(module, clz, name, path):
    # Create a service instance
    svc = get_instance(module, clz)

    # Package the pretrained model artifact
    svc.pack(name, path)

    # Save the service to the model registry for serving
    saved_path = svc.save()

    print('Saved model to ' + saved_path)


def get_instance(module_name, class_name):
    module = __import__(module_name)
    class_ = getattr(module, class_name)
    return class_()


yatai_proto = LazyLoader('yatai_proto', globals(), 'bentoml.yatai.proto')


@click.command()
@click.option('--bento', help='bento service name')
@click.option('--registry', help='registry name')
def deploy_to_knative(bento, registry):
    # Get saved path
    if ':' in bento:
        name, version = bento.split(':')
    else:
        name = bento
        version = 'latest'
    
    yatai_client = get_default_yatai_client()
    result = yatai_client.repository.get(name, version)
    if result.status.status_code != yatai_proto.status_pb2.Status.OK:
        error_code, error_message = status_pb_to_error_code_and_message(result.status)
        print(f'{error_code}:{error_message}')
        exit 1
    
    saved_path = result.bento.uri.uri

    # Build Docker image
    client = docker.from_env()
    tag = f'{registry}/{name}:{version}'
    client.build(path=saved_path, tag=tag)
    for line in client.push(tag, stream=True, decode=True):
        print(line)
    
    # Generate KNative manifest
    output_dir = tempfile.TemporaryDirectory(dir='/tmp')
    root_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    resources_dir = os.path.join(root_dir, 'templates/knative')
    env = Environment(loader=FileSystemLoader(resources_dir))
    template = env.get_template('service.yaml')
    svc_name = stringcase.spinalcase(name)
    yaml_file = os.path.join(output_dir, 'service.yaml')
    template.stream(name=svc_name, registry=registry).dump(yaml_file)

    # Deploy to KNative
    from_yaml(yaml_file)

    # Cleanup
    output_dir.cleanup()


def get_default_yatai_client():
    from bentoml.yatai.client import YataiClient

    return YataiClient()


# This function assumes the status is not status.OK
def status_pb_to_error_code_and_message(pb_status) -> (int, str):
    from bentoml.yatai.proto import status_pb2

    assert pb_status.status_code != status_pb2.Status.OK
    error_code = status_pb2.Status.Code.Name(pb_status.status_code)
    error_message = pb_status.error_message
    return error_code, error_message



#!/usr/bin/env python3
# This utility attempts to provide functionality similar to `kubectl apply -f`
# - load and parse yaml
# - try to figure out the object type and API to use
# - figure out if the resource already exists, in which case
#   it needs to be patched or replaced alltogether.
# - otherwise just create a new resource.
# See https://stackoverflow.com/questions/36307950/kubernetes-api-call-equivalent-to-kubectl-apply

import kubernetes
import logging
import re
import sys
import yaml


def from_yaml(raw, client=None, **kwargs):
    ''' Invoke the K8s API to create or replace an object given a YAML spec.

        @param raw: either a string or an open input stream with a YAML formatted 
                    spec, as you would use for `kubectl apply -f`
        @param client: (optional) preconfigured client environment to use for invocation
        @param kwargs: (optional) further arguments to pass to the create/replace call
        @return: response object from the Kubernetes API call
    '''
    for obj in yaml.load_all(raw, Loader=yaml.FullLoader):
        create_update_or_replace(obj, client, **kwargs)


def create_update_or_replace(obj, client=None, **kwargs):
    ''' Invoke the K8s API to create or replace a Kubernetes object.

        The first attempt is to create (insert) this object. When this is rejected because
        of an existing object with same name, we attempt to patch the existing object.
        As a last resort, if even the patch is rejected, we *delete* the existing object
        and recreate from scratch.

        @param obj: complete object specification including API version and metadata.
        @param client: (optional) preconfigured client environment to use for invocation
        @param kwargs: (optional) further arguments to pass to the create/replace call
        @return: response object from the Kubernetes API call
    '''
    api_instance = find_api_instance(obj, client)
    try:
        res = invoke_api(api_instance, 'create', obj, **kwargs)
        logging.debug('K8s: %s created -> uid=%s', describe(obj), res.metadata.uid)
    except kubernetes.client.rest.ApiException as e:
        if e.reason != 'Conflict':
            raise

        try:
            # asking for forgiveness...
            res = invoke_api(api_instance, 'patch', obj, **kwargs)
            logging.debug('K8s: %s PATCHED -> uid=%s', describe(obj), res.metadata.uid)
        except kubernetes.client.rest.ApiException as e:
            if e.reason != 'Unprocessable Entity':
                raise

            try:
                # second attempt... delete the existing object and re-insert
                logging.debug('K8s: replacing %s FAILED. Attempting deletion and recreation...', describe(obj))
                res = invoke_api(api_instance, 'delete', obj, **kwargs)
                logging.debug('K8s: %s DELETED...', describe(obj))
                res = invoke_api(api_instance, 'create', obj, **kwargs)
                logging.debug('K8s: %s CREATED -> uid=%s', describe(obj), res.metadata.uid)
            except Exception as ex:
                message = 'K8s: FAILURE updating %s. Exception: %s' % (describe(obj), ex)
                logging.error(message)
                raise RuntimeError(message)
    
    return res


def patch_object(obj, client=None, **kwargs):
    api_instance = find_api_instance(obj, client)
    try:
        res = invoke_api(api_instance, 'patch', obj, **kwargs)
        logging.debug('K8s: %s PATCHED -> uid=%s', describe(obj), res.metadata.uid)
        return res
    except kubernetes.client.rest.ApiException as e:
        if e.reason == 'Unprocessable Entity':
            message = 'K8s: patch for %s rejected. Exception: %s' % (describe(obj), e)
            logging.error(message)
            raise RuntimeError(message)
        else:
            raise


def delete_object(obj, client=None, **kwargs):
    api_instance = find_api_instance(obj, client)
    try:
        res = invoke_api(api_instance, 'delete', obj, **kwargs)
        logging.debug('K8s: %s DELETED. uid was: %s', describe(obj), res.details and res.details.uid or '?')
        return True
    except kubernetes.client.rest.ApiException as e:
        if e.reason == 'Not Found':
            logging.warning('K8s: %s does not exist (anymore).', describe(obj))
            return False
        else:
            message = 'K8s: deleting %s FAILED. Exception: %s' % (describe(obj), e)
            logging.error(message)
            raise RuntimeError(message)


def find_api_instance(obj, client=None):
    ''' Investigate the object spec and lookup the corresponding API object.
    
        @param client: (optional) preconfigured client environment to use for invocation
        @return: a client instance wired to the appropriate API
    '''
    grp, _, ver = obj['apiVersion'].partition('/')
    if ver == '':
        ver = grp
        grp = 'core'

    # Strip 'k8s.io', camel-case-join dot separated parts. e.g. rbac.authorization.k8s.io -> RbacAuthorzation
    grp = ''.join(part.capitalize() for part in grp.rsplit('.k8s.io', 1)[0].split('.'))
    ver = ver.capitalize()

    api_instance = '%s%sApi' % (grp, ver)
    return getattr(kubernetes.client, api_instance)(client)


def invoke_api(api_instance, action, obj, **args):
    ''' Find a suitable function and perform the actual API invocation.

        @param api_instance: client object for the invocation, wired to correct API version
        @param action: either 'create' (to inject a new objet) or 'replace','patch','delete'
        @param obj: the full object spec to be passed into the API invocation
        @param args: (optional) extraneous arguments to pass
        @return: response object from the Kubernetes API call
    '''
    # transform ActionType from yaml into action_type for swagger API
    kind = camel_to_snake(obj['kind'])

    # determine namespace to place the object in, supply default
    try:
        namespace = obj['metadata']['namespace']
    except:
        namespace = 'default'

    fn_name = '%s_%s' %(action,kind)
    if hasattr(api_instance, fn_name):
        # namespace agnostic API
        fn = getattr(api_instance, fn_name)
    else:
        fn_name = '%s_namespaced_%s' %(action,kind)
        fn = getattr(api_instance, fn_name)
        args['namespace'] = namespace

    if not 'create' in fn_name:
        args['name'] = obj['metadata']['name']
    
    if 'delete' in fn_name:
        from kubernetes.client.models.v1_delete_options import V1DeleteOptions
        obj = V1DeleteOptions()

    return fn(body=obj, **args)


def describe(obj):
    return "%s '%s'" % (obj['kind'], obj['metadata']['name'])


def camel_to_snake(str):
    str = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', str)
    str = re.sub('([a-z0-9])([A-Z])', r'\1_\2', str).lower()
    return str


if __name__ == '__main__':
    from_yaml(sys.argv[0])



import setuptools

with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name='bentoutils',
    version='0.0.3',
    author='Mark Moloney',
    author_email='m4rkmo@gmail.com',
    description='Utilities for working with BentoML',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/markmo/bentoutils',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    install_requires=[
        'BentoML==0.9.0',
        'click==7.1.2',
        'docker==4.3.1',
        'Jinja2==2.11.2',
        'kubernetes==11.0.0',
        'PyYAML==5.3.1',
        'simpletransformers==0.28.2',
        'stringcase==1.2.0',
        'torch==1.6.0+cpu',
        'torchvision==0.7.0+cpu',
        'transformers==2.10.0',
    ],
    dependency_links=[
        'https://download.pytorch.org/whl/torch_stable.html',
    ],
    entry_points='''
        [console_scripts]
        bentopack=bentoutils.cli:pack
        deploy_to_knative=bentoutils.cli:deploy_to_knative
    ''',
    python_requires='>=3.6',
)
